#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
# pylint: disable=R0913,C0301,R0917
"""
Restate Context
"""

import abc
from random import Random
from uuid import UUID
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Dict, List, Optional, TypeVar, Union, Coroutine, overload, ParamSpec
import typing
from datetime import timedelta

from restate.serde import DefaultSerde, Serde

T = TypeVar("T")
I = TypeVar("I")
O = TypeVar("O")
P = ParamSpec("P")

HandlerType = Union[Callable[[Any, I], Awaitable[O]], Callable[[Any], Awaitable[O]]]
RunAction = Union[Callable[..., Coroutine[Any, Any, T]], Callable[..., T]]


# pylint: disable=R0902
@dataclass
class RunOptions(typing.Generic[T]):
    """
    Options for running an action.
    """

    serde: Serde[T] = DefaultSerde()
    """The serialization/deserialization mechanism. - if the default serde is used, a default serializer will be used based on the type.
                    See also 'type_hint'."""
    type_hint: Optional[typing.Type[T]] = None
    """The type hint of the return value of the action. This is used to pick the serializer. If None, the type hint will be inferred from the action's return type, or the provided serializer."""
    max_attempts: Optional[int] = None
    """Max number of attempts (including the initial), before giving up.

    When giving up, `ctx.run` will throw a `TerminalError` wrapping the original error message."""
    max_duration: Optional[timedelta] = None
    """Max duration of retries, before giving up.

    When giving up, `ctx.run` will throw a `TerminalError` wrapping the original error message."""
    initial_retry_interval: Optional[timedelta] = None
    """Initial interval for the first retry attempt.
    Retry interval will grow by a factor specified in `retry_interval_factor`.

    If any of the other retry related fields is specified, the default for this field is 50 milliseconds, otherwise restate will fallback to the overall invocation retry policy."""
    max_retry_interval: Optional[timedelta] = None
    """Max interval between retries.
    Retry interval will grow by a factor specified in `retry_interval_factor`.

    The default is 10 seconds."""
    retry_interval_factor: Optional[float] = None
    """Exponentiation factor to use when computing the next retry delay.

    If any of the other retry related fields is specified, the default for this field is `2`, meaning retry interval will double at each attempt, otherwise restate will fallback to the overall invocation retry policy."""
    max_retry_duration: Optional[timedelta] = None
    """Deprecated: Use max_duration instead."""


# pylint: disable=R0903
class RestateDurableFuture(typing.Generic[T], Awaitable[T]):
    """
    Represents a durable future.
    """

    @abc.abstractmethod
    def __await__(self) -> typing.Generator[Any, Any, T]:
        pass


# pylint: disable=R0903
class RestateDurableCallFuture(RestateDurableFuture[T]):
    """
    Represents a durable call future.
    """

    @abc.abstractmethod
    async def invocation_id(self) -> str:
        """
        Returns the invocation id of the call.
        """

    @abc.abstractmethod
    async def cancel_invocation(self) -> None:
        """
        Cancels the invocation.

        Just a utility shortcut to:
        .. code-block:: python

            await ctx.cancel_invocation(await f.invocation_id())
        """


class RestateDurableSleepFuture(RestateDurableFuture[None]):
    """
    Represents a durable sleep future.
    """

    @abc.abstractmethod
    def __await__(self) -> typing.Generator[Any, Any, None]:
        pass


class AttemptFinishedEvent(abc.ABC):
    """
    Represents an attempt finished event.

    This event is used to signal that an attempt has finished (either successfully or with an error), and it is now
    safe to cleanup any attempt related resources, such as pending ctx.run() 3rd party calls, or any other resources that
    are only valid for the duration of the attempt.

    An attempt is considered finished when either the connection to the restate server is closed, the invocation is completed, or a transient
    error occurs.
    """

    @abc.abstractmethod
    def is_set(self) -> bool:
        """
        Returns True if the event is set, False otherwise.
        """

    @abc.abstractmethod
    async def wait(self):
        """
        Waits for the event to be set.
        """


@dataclass
class Request:
    """
    Represents an ingress request.

    Attributes:
        id (str): The unique identifier of the request.
        headers (dict[str, str]): The headers of the request.
        attempt_headers (dict[str, str]): The attempt headers of the request.
        body (bytes): The body of the request.
        attempt_finished_event (AttemptFinishedEvent): The teardown event of the request.
    """

    id: str
    headers: Dict[str, str]
    attempt_headers: Dict[str, str]
    body: bytes
    attempt_finished_event: AttemptFinishedEvent


class KeyValueStore(abc.ABC):
    """
    A key scoped key-value store.

    This class defines the interface for a key-value store,
    which allows storing and retrieving values
    based on a unique key.

    """

    @abc.abstractmethod
    def get(
        self, name: str, serde: Serde[T] = DefaultSerde(), type_hint: Optional[typing.Type[T]] = None
    ) -> Awaitable[Optional[T]]:
        """
        Retrieves the value associated with the given name.

        Args:
            name: The state name
            serde: The serialization/deserialization mechanism. - if the default serde is used, a default serializer will be used based on the type.
                    See also 'type_hint'.
            type_hint: The type hint of the return value. This is used to pick the serializer. If None, the type hint will be inferred from the action's return type, or the provided serializer.
        """

    @abc.abstractmethod
    def state_keys(self) -> Awaitable[List[str]]:
        """Returns the list of keys in the store."""

    @abc.abstractmethod
    def set(self, name: str, value: T, serde: Serde[T] = DefaultSerde()) -> None:
        """set the value associated with the given name."""

    @abc.abstractmethod
    def clear(self, name: str) -> None:
        """clear the value associated with the given name."""

    @abc.abstractmethod
    def clear_all(self) -> None:
        """clear all the values in the store."""


# pylint: disable=R0903
class SendHandle(abc.ABC):
    """
    Represents a send operation.
    """

    @abc.abstractmethod
    async def invocation_id(self) -> str:
        """
        Returns the invocation id of the send operation.
        """

    @abc.abstractmethod
    async def cancel_invocation(self) -> None:
        """
        Cancels the invocation.

        Just a utility shortcut to:
        .. code-block:: python

            await ctx.cancel_invocation(await f.invocation_id())
        """


class Context(abc.ABC):
    """
    Represents the context of the current invocation.
    """

    @abc.abstractmethod
    def request(self) -> Request:
        """
        Returns the request object.
        """

    @abc.abstractmethod
    def random(self) -> Random:
        """
        Returns a Random instance inherently predictable, deterministically seeded by Restate.

        This instance is useful to generate identifiers, idempotency keys, and for uniform sampling from a set of options.
        """

    @abc.abstractmethod
    def uuid(self) -> UUID:
        """
        Returns a random UUID, deterministically seeded.

        This UUID will be stable across retries and replays.
        """

    @abc.abstractmethod
    def time(self) -> RestateDurableFuture[float]:
        """
        Returns the result of time.time(), durably recorded in the journal.

        This timestamp will be stable across retries and replays.
        """

    @overload
    @abc.abstractmethod
    def run(
        self,
        name: str,
        action: Callable[..., Coroutine[Any, Any, T]],
        serde: Serde[T] = DefaultSerde(),
        max_attempts: typing.Optional[int] = None,
        max_retry_duration: typing.Optional[timedelta] = None,
        type_hint: Optional[typing.Type[T]] = None,
        args: Optional[typing.Tuple[Any, ...]] = None,
    ) -> RestateDurableFuture[T]:
        """
        Runs the given action with the given name.

        DEPRECATED: Use ctx.run_typed instead.

        Args:
            name: The name of the action.
            action: The action to run.
            serde: The serialization/deserialization mechanism. - if the default serde is used, a default serializer will be used based on the type.
                    See also 'type_hint'.
            max_attempts:   The maximum number of retry attempts to complete the action.
                            If None, the action will be retried indefinitely, until it succeeds.
                            Otherwise, the action will be retried until the maximum number of attempts is reached and then it will raise a TerminalError.
            max_retry_duration: The maximum duration for retrying. If None, the action will be retried indefinitely, until it succeeds.
                                Otherwise, the action will be retried until the maximum duration is reached and then it will raise a TerminalError.
            type_hint: The type hint of the return value of the action.
                        This is used to pick the serializer. If None, the type hint will be inferred from the action's return type, or the provided serializer.

        """

    @overload
    @abc.abstractmethod
    def run(
        self,
        name: str,
        action: Callable[..., T],
        serde: Serde[T] = DefaultSerde(),
        max_attempts: typing.Optional[int] = None,
        max_retry_duration: typing.Optional[timedelta] = None,
        type_hint: Optional[typing.Type[T]] = None,
        args: Optional[typing.Tuple[Any, ...]] = None,
    ) -> RestateDurableFuture[T]:
        """
        Runs the given coroutine action with the given name.

        DEPRECATED: Use ctx.run_typed instead.

        Args:
            name: The name of the action.
            action: The action to run.
            serde: The serialization/deserialization mechanism. - if the default serde is used, a default serializer will be used based on the type.
                    See also 'type_hint'.
            max_attempts:   The maximum number of retry attempts to complete the action.
                            If None, the action will be retried indefinitely, until it succeeds.
                            Otherwise, the action will be retried until the maximum number of attempts is reached and then it will raise a TerminalError.
            max_retry_duration: The maximum duration for retrying. If None, the action will be retried indefinitely, until it succeeds.
                                Otherwise, the action will be retried until the maximum duration is reached and then it will raise a TerminalError.
            type_hint: The type hint of the return value of the action.
                        This is used to pick the serializer. If None, the type hint will be inferred from the action's return type, or the provided serializer.

        """

    @abc.abstractmethod
    def run(
        self,
        name: str,
        action: RunAction[T],
        serde: Serde[T] = DefaultSerde(),
        max_attempts: typing.Optional[int] = None,
        max_retry_duration: typing.Optional[timedelta] = None,
        type_hint: Optional[typing.Type[T]] = None,
        args: Optional[typing.Tuple[Any, ...]] = None,
    ) -> RestateDurableFuture[T]:
        """
        Runs the given action with the given name.

        DEPRECATED: Use ctx.run_typed instead.

        Args:
            name: The name of the action.
            action: The action to run.
            serde: The serialization/deserialization mechanism. - if the default serde is used, a default serializer will be used based on the type.
                    See also 'type_hint'.
            max_attempts:   The maximum number of retry attempts to complete the action.
                            If None, the action will be retried indefinitely, until it succeeds.
                            Otherwise, the action will be retried until the maximum number of attempts is reached and then it will raise a TerminalError.
            max_retry_duration: The maximum duration for retrying. If None, the action will be retried indefinitely, until it succeeds.
                                Otherwise, the action will be retried until the maximum duration is reached and then it will raise a TerminalError.
            type_hint: The type hint of the return value of the action.
                        This is used to pick the serializer. If None, the type hint will be inferred from the action's return type, or the provided serializer.

        """

    @overload
    @abc.abstractmethod
    def run_typed(
        self,
        name: str,
        action: Callable[P, Coroutine[Any, Any, T]],
        options: RunOptions[T] = RunOptions(),
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RestateDurableFuture[T]:
        """
        Typed version of run that provides type hints for the function arguments.
        Runs the given action with the given name.

        Args:
            name: The name of the action.
            action: The action to run.
            options: The options for the run.
            *args: The arguments to pass to the action.
            **kwargs: The keyword arguments to pass to the action.
        """

    @overload
    @abc.abstractmethod
    def run_typed(
        self,
        name: str,
        action: Callable[P, T],
        options: RunOptions[T] = RunOptions(),
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RestateDurableFuture[T]:
        """
        Typed version of run that provides type hints for the function arguments.
        Runs the given coroutine action with the given name.

        Args:
            name: The name of the action.
            action: The action to run.
            options: The options for the run.
            *args: The arguments to pass to the action.
            **kwargs: The keyword arguments to pass to the action.
        """

    @abc.abstractmethod
    def run_typed(
        self,
        name: str,
        action: Union[Callable[P, Coroutine[Any, Any, T]], Callable[P, T]],
        options: RunOptions[T] = RunOptions(),
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RestateDurableFuture[T]:
        """
        Typed version of run that provides type hints for the function arguments.
        Runs the given action with the given name.

        Args:
            name: The name of the action.
            action: The action to run.
            options: The options for the run.
            *args: The arguments to pass to the action.
            **kwargs: The keyword arguments to pass to the action.

        """

    @abc.abstractmethod
    def sleep(self, delta: timedelta, name: Optional[str] = None) -> RestateDurableSleepFuture:
        """
        Suspends the current invocation for the given duration
        """

    @abc.abstractmethod
    def service_call(
        self,
        tpe: HandlerType[I, O],
        arg: I,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateDurableCallFuture[O]:
        """
        Invokes the given service with the given argument.
        """

    @abc.abstractmethod
    def service_send(
        self,
        tpe: HandlerType[I, O],
        arg: I,
        send_delay: Optional[timedelta] = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> SendHandle:
        """
        Invokes the given service with the given argument.
        """

    @abc.abstractmethod
    def object_call(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateDurableCallFuture[O]:
        """
        Invokes the given object with the given argument.
        """

    @abc.abstractmethod
    def object_send(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        send_delay: Optional[timedelta] = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> SendHandle:
        """
        Send a message to an object with the given argument.
        """

    @abc.abstractmethod
    def workflow_call(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateDurableCallFuture[O]:
        """
        Invokes the given workflow with the given argument.
        """

    @abc.abstractmethod
    def workflow_send(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        send_delay: Optional[timedelta] = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> SendHandle:
        """
        Send a message to an object with the given argument.
        """

    # pylint: disable=R0913
    @abc.abstractmethod
    def generic_call(
        self,
        service: str,
        handler: str,
        arg: bytes,
        key: Optional[str] = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateDurableCallFuture[bytes]:
        """
        Invokes the given generic service/handler with the given argument.
        """

    @abc.abstractmethod
    def generic_send(
        self,
        service: str,
        handler: str,
        arg: bytes,
        key: Optional[str] = None,
        send_delay: Optional[timedelta] = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> SendHandle:
        """
        Send a message to a generic service/handler with the given argument.
        """

    @abc.abstractmethod
    def awakeable(
        self, serde: Serde[T] = DefaultSerde(), type_hint: Optional[typing.Type[T]] = None
    ) -> typing.Tuple[str, RestateDurableFuture[T]]:
        """
        Returns the name of the awakeable and the future to be awaited.
        """

    @abc.abstractmethod
    def resolve_awakeable(self, name: str, value: I, serde: Serde[I] = DefaultSerde()) -> None:
        """
        Resolves the awakeable with the given name.
        """

    @abc.abstractmethod
    def reject_awakeable(self, name: str, failure_message: str, failure_code: int = 500) -> None:
        """
        Rejects the awakeable with the given name.
        """

    @abc.abstractmethod
    def cancel_invocation(self, invocation_id: str):
        """
        Cancels the invocation with the given id.
        """

    @abc.abstractmethod
    def attach_invocation(
        self, invocation_id: str, serde: Serde[T] = DefaultSerde(), type_hint: typing.Optional[typing.Type[T]] = None
    ) -> RestateDurableFuture[T]:
        """
        Attaches the invocation with the given id.
        """


class ObjectContext(Context, KeyValueStore):
    """
    Represents the context of the current invocation.
    """

    @abc.abstractmethod
    def key(self) -> str:
        """
        Returns the key of the current object.
        """


class ObjectSharedContext(Context):
    """
    Represents the context of the current invocation.
    """

    @abc.abstractmethod
    def key(self) -> str:
        """Returns the key of the current object."""

    @abc.abstractmethod
    def get(
        self, name: str, serde: Serde[T] = DefaultSerde(), type_hint: Optional[typing.Type[T]] = None
    ) -> RestateDurableFuture[Optional[T]]:
        """
        Retrieves the value associated with the given name.

        Args:
            name: The state name
            serde: The serialization/deserialization mechanism. - if the default serde is used, a default serializer will be used based on the type.
                    See also 'type_hint'.
            type_hint: The type hint of the return value. This is used to pick the serializer. If None, the type hint will be inferred from the action's return type, or the provided serializer.
        """

    @abc.abstractmethod
    def state_keys(self) -> Awaitable[List[str]]:
        """
        Returns the list of keys in the store.
        """


class DurablePromise(typing.Generic[T]):
    """
    Represents a durable promise.
    """

    def __init__(self, name: str, serde: Serde[T] = DefaultSerde()) -> None:
        self.name = name
        self.serde = serde

    @abc.abstractmethod
    def resolve(self, value: T) -> Awaitable[None]:
        """
        Resolves the promise with the given value.
        """

    @abc.abstractmethod
    def reject(self, message: str, code: int = 500) -> Awaitable[None]:
        """
        Rejects the promise with the given message and code.
        """

    @abc.abstractmethod
    def peek(self) -> Awaitable[typing.Optional[T]]:
        """
        Returns the value of the promise if it is resolved, None otherwise.
        """

    @abc.abstractmethod
    def value(self) -> RestateDurableFuture[T]:
        """
        Returns the value of the promise if it is resolved, None otherwise.
        """

    @abc.abstractmethod
    def __await__(self) -> typing.Generator[Any, Any, T]:
        """
        Returns the value of the promise. This is a shortcut for calling value() and awaiting it.
        """


class WorkflowContext(ObjectContext):
    """
    Represents the context of the current workflow invocation.
    """

    @abc.abstractmethod
    def promise(
        self, name: str, serde: Serde[T] = DefaultSerde(), type_hint: Optional[typing.Type[T]] = None
    ) -> DurablePromise[T]:
        """
        Returns a durable promise with the given name.
        """


class WorkflowSharedContext(ObjectSharedContext):
    """
    Represents the context of the current workflow invocation.
    """

    @abc.abstractmethod
    def promise(
        self, name: str, serde: Serde[T] = DefaultSerde(), type_hint: Optional[typing.Type[T]] = None
    ) -> DurablePromise[T]:
        """
        Returns a durable promise with the given name.
        """



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

import os
from datetime import timedelta
from typing import Iterable, List, Union, TypedDict, Literal, Any
from restate import VirtualObject, ObjectSharedContext, ObjectContext, RestateDurableFuture, RestateDurableSleepFuture
from restate import select, wait_completed, as_completed
from restate.exceptions import TerminalError

virtual_object_command_interpreter = VirtualObject("VirtualObjectCommandInterpreter")


@virtual_object_command_interpreter.handler(name="getResults", kind="shared")
async def get_results(ctx: ObjectSharedContext | ObjectContext) -> List[str]:
    return (await ctx.get("results")) or []


@virtual_object_command_interpreter.handler(name="hasAwakeable", kind="shared")
async def has_awakeable(ctx: ObjectSharedContext, awk_key: str) -> bool:
    awk_id = await ctx.get("awk-" + awk_key)
    if awk_id:
        return True
    return False


class CreateAwakeable(TypedDict):
    type: Literal["createAwakeable"]
    awakeableKey: str


class Sleep(TypedDict):
    type: Literal["sleep"]
    timeoutMillis: int


class RunThrowTerminalException(TypedDict):
    type: Literal["runThrowTerminalException"]
    reason: str


AwaitableCommand = Union[CreateAwakeable, Sleep, RunThrowTerminalException]


class AwaitOne(TypedDict):
    type: Literal["awaitOne"]
    command: AwaitableCommand


class AwaitAnySuccessful(TypedDict):
    type: Literal["awaitAnySuccessful"]
    commands: List[AwaitableCommand]


class AwaitAny(TypedDict):
    type: Literal["awaitAny"]
    commands: List[AwaitableCommand]


class AwaitAwakeableOrTimeout(TypedDict):
    type: Literal["awaitAwakeableOrTimeout"]
    awakeableKey: str
    timeoutMillis: int


class ResolveAwakeable(TypedDict):
    type: Literal["resolveAwakeable"]
    awakeableKey: str
    value: str


class RejectAwakeable(TypedDict):
    type: Literal["rejectAwakeable"]
    awakeableKey: str
    reason: str


class GetEnvVariable(TypedDict):
    type: Literal["getEnvVariable"]
    envName: str


Command = Union[
    AwaitOne, AwaitAny, AwaitAnySuccessful, AwaitAwakeableOrTimeout, ResolveAwakeable, RejectAwakeable, GetEnvVariable
]


class InterpretRequest(TypedDict):
    commands: Iterable[Command]


@virtual_object_command_interpreter.handler(name="resolveAwakeable", kind="shared")
async def resolve_awakeable(ctx: ObjectSharedContext | ObjectContext, req: ResolveAwakeable):
    awk_id = await ctx.get("awk-" + req["awakeableKey"])
    if not awk_id:
        raise TerminalError(message="No awakeable is registered")
    ctx.resolve_awakeable(awk_id, req["value"])


@virtual_object_command_interpreter.handler(name="rejectAwakeable", kind="shared")
async def reject_awakeable(ctx: ObjectSharedContext | ObjectContext, req: RejectAwakeable):
    awk_id = await ctx.get("awk-" + req["awakeableKey"])
    if not awk_id:
        raise TerminalError(message="No awakeable is registered")
    ctx.reject_awakeable(awk_id, req["reason"])


def to_durable_future(ctx: ObjectContext, cmd: AwaitableCommand) -> RestateDurableFuture[Any]:
    if cmd["type"] == "createAwakeable":
        awk_id, awakeable = ctx.awakeable()
        ctx.set("awk-" + cmd["awakeableKey"], awk_id)
        return awakeable
    elif cmd["type"] == "sleep":
        return ctx.sleep(timedelta(milliseconds=cmd["timeoutMillis"]))
    elif cmd["type"] == "runThrowTerminalException":

        def side_effect(reason: str):
            raise TerminalError(message=reason)

        res = ctx.run_typed("run should fail command", side_effect, reason=cmd["reason"])
        return res


@virtual_object_command_interpreter.handler(name="interpretCommands")
async def interpret_commands(ctx: ObjectContext, req: InterpretRequest):
    result = ""

    for cmd in req["commands"]:
        if cmd["type"] == "awaitAwakeableOrTimeout":
            awk_id, awakeable = ctx.awakeable()
            ctx.set("awk-" + cmd["awakeableKey"], awk_id)
            match await select(awakeable=awakeable, timeout=ctx.sleep(timedelta(milliseconds=cmd["timeoutMillis"]))):
                case ["awakeable", awk_res]:
                    result = awk_res
                case ["timeout", _]:
                    raise TerminalError(message="await-timeout", status_code=500)
        elif cmd["type"] == "resolveAwakeable":
            await resolve_awakeable(ctx, cmd)
            result = ""
        elif cmd["type"] == "rejectAwakeable":
            await reject_awakeable(ctx, cmd)
            result = ""
        elif cmd["type"] == "getEnvVariable":
            env_name = cmd["envName"]

            def side_effect(env_name: str):
                return os.environ.get(env_name, "")

            result = await ctx.run_typed("get_env", side_effect, env_name=env_name)
        elif cmd["type"] == "awaitOne":
            awaitable = to_durable_future(ctx, cmd["command"])
            # We need this dance because the Python SDK doesn't support .map on futures
            if isinstance(awaitable, RestateDurableSleepFuture):
                await awaitable
                result = "sleep"
            else:
                result = await awaitable
        elif cmd["type"] == "awaitAny":
            futures = [to_durable_future(ctx, c) for c in cmd["commands"]]
            done, _ = await wait_completed(*futures)
            done_fut = done[0]
            # We need this dance because the Python SDK doesn't support .map on futures
            if isinstance(done_fut, RestateDurableSleepFuture):
                await done_fut
                result = "sleep"
            else:
                result = await done_fut
        elif cmd["type"] == "awaitAnySuccessful":
            futures = [to_durable_future(ctx, c) for c in cmd["commands"]]
            async for done_fut in as_completed(*futures):
                try:
                    # We need this dance because the Python SDK doesn't support .map on futures
                    if isinstance(done_fut, RestateDurableSleepFuture):
                        await done_fut
                        result = "sleep"
                        break
                    result = await done_fut
                    break
                except TerminalError:
                    pass

        last_results = await get_results(ctx)
        last_results.append(result)
        ctx.set("results", last_results)

    return result



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613


#
# uv run examples/example.py
#

import logging
import restate

from greeter import greeter
from random_greeter import random_greeter
from virtual_object import counter
from workflow import payment
from pydantic_greeter import pydantic_greeter
from concurrent_greeter import concurrent_greeter

logging.basicConfig(level=logging.INFO)

app = restate.app(services=[greeter, random_greeter, counter, payment, pydantic_greeter, concurrent_greeter])

if __name__ == "__main__":
    import hypercorn
    import hypercorn.asyncio
    import asyncio

    conf = hypercorn.Config()
    conf.bind = ["0.0.0.0:9080"]
    asyncio.run(hypercorn.asyncio.serve(app, conf))



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
# pylint: disable=R0913,C0301,R0917
# pylint: disable=line-too-long
"""combines multiple futures into a single future"""

from typing import Any, List, Tuple
from restate.exceptions import TerminalError
from restate.context import RestateDurableFuture
from restate.server_context import ServerDurableFuture, ServerInvocationContext


async def gather(*futures: RestateDurableFuture[Any]) -> List[RestateDurableFuture[Any]]:
    """
    Blocks until all futures are completed.

    Returns a list of all futures.
    """
    async for _ in as_completed(*futures):
        pass
    return list(futures)


async def select(**kws: RestateDurableFuture[Any]) -> List[Any]:
    """
    Blocks until one of the futures is completed.

    Example:

    who, what = await select(car=f1, hotel=f2, flight=f3)
    if who == "car":
        print(what)
    elif who == "hotel":
        print(what)
    elif who == "flight":
        print(what)

    works the best with matching:

    match await select(result=ctx.promise("verify.payment"), timeout=ctx.sleep(timedelta(seconds=10))):
    case ['result', "approved"]:
        return { "success" : True }
    case ['result', "declined"]:
        raise TerminalError(message="Payment declined", status_code=401)
    case ['timeout', _]:
        raise TerminalError(message="Payment verification timed out", status_code=410)

    """
    if not kws:
        raise ValueError("At least one future must be passed.")
    reverse = {f: key for key, f in kws.items()}
    async for f in as_completed(*kws.values()):
        return [reverse[f], await f]
    assert False, "unreachable"


async def as_completed(*futures: RestateDurableFuture[Any]):
    """
    Returns an iterator that yields the futures as they are completed.

    example:

    async for future in as_completed(f1, f2, f3):
        # do something with the completed future
        print(await future)  # prints the result of the future

    """
    remaining = list(futures)
    while remaining:
        completed, waiting = await wait_completed(*remaining)
        for f in completed:
            yield f
        remaining = waiting


async def wait_completed(
    *args: RestateDurableFuture[Any],
) -> Tuple[List[RestateDurableFuture[Any]], List[RestateDurableFuture[Any]]]:
    """
    Blocks until at least one of the futures is completed.

    Returns a tuple of two lists: the first list contains the futures that are completed,
    the second list contains the futures that are not completed.
    """
    handles: List[int] = []
    context: ServerInvocationContext | None = None
    completed = []
    uncompleted = []
    futures = list(args)

    if not futures:
        return [], []
    for f in futures:
        if not isinstance(f, ServerDurableFuture):
            raise TerminalError("All futures must SDK created futures.")
        if context is None:
            context = f.context
        elif context is not f.context:
            raise TerminalError("All futures must be created by the same SDK context.")
        if f.is_completed():
            completed.append(f)
        else:
            handles.append(f.handle)
            uncompleted.append(f)

    if completed:
        # the user had passed some completed futures, so we can return them immediately
        return completed, uncompleted  # type: ignore

    completed = []
    uncompleted = []
    assert context is not None
    await context.create_poll_or_cancel_coroutine(handles)

    for index, handle in enumerate(handles):
        future = futures[index]
        if context.vm.is_completed(handle):
            completed.append(future)  # type: ignore
        else:
            uncompleted.append(future)  # type: ignore
    return completed, uncompleted  # type: ignore



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#

# pylint: disable=R0917
# pylint: disable=R0801
"""
This module defines the Service class for representing a restate service.
"""

from functools import wraps
import inspect
import typing
from datetime import timedelta

from restate.retry_policy import InvocationRetryPolicy
from restate.serde import DefaultSerde, Serde
from restate.handler import Handler, HandlerIO, ServiceTag, make_handler

I = typing.TypeVar("I")
O = typing.TypeVar("O")
T = typing.TypeVar("T")


# disable too many arguments warning
# pylint: disable=R0913

# disable line too long warning
# pylint: disable=C0301

# disable similar lines warning
# pylint: disable=R0801


# pylint: disable=R0902
class Workflow:
    """
    Represents a restate workflow.

    Args:
        name (str): The name of the workflow.
        description (str, optional): Documentation as shown in the UI, Admin REST API, and the generated OpenAPI documentation of this service.
        metadata (Dict[str, str], optional): Service metadata, as propagated in the Admin REST API.
        inactivity_timeout (timedelta, optional): This timer guards against stalled invocations. Once it expires, Restate triggers a graceful
            termination by asking the invocation to suspend (which preserves intermediate progress).
            The abort_timeout is used to abort the invocation, in case it doesn't react to the request to suspend.
            This overrides the default inactivity timeout configured in the restate-server for all invocations to this service.
            NOTE: You can set this field only if you register this service against restate-server >= 1.4,
            otherwise the service discovery will fail.
        abort_timeout (timedelta, optional): This timer guards against stalled service/handler invocations that are supposed to terminate. The
            abort timeout is started after the inactivity_timeout has expired and the service/handler invocation has been asked to gracefully terminate.
            Once the timer expires, it will abort the service/handler invocation.
            This timer potentially *interrupts* user code. If the user code needs longer to gracefully terminate, then this value needs to be set accordingly.
            This overrides the default abort timeout configured in the restate-server for all invocations to this service.
            NOTE: You can set this field only if you register this service against restate-server >= 1.4,
            otherwise the service discovery will fail.
        journal_retention (timedelta, optional): The journal retention. When set, this applies to all requests to all handlers of this service.
            In case the request has an idempotency key, the idempotency_retention caps the journal retention time.
            NOTE: You can set this field only if you register this service against restate-server >= 1.4,
            otherwise the service discovery will fail.
        idempotency_retention (timedelta, optional): The retention duration of idempotent requests to this service.
            NOTE: You can set this field only if you register this service against restate-server >= 1.4,
            otherwise the service discovery will fail.
        enable_lazy_state (bool, optional): When set to True, lazy state will be enabled for all invocations to this service.
            NOTE: You can set this field only if you register this service against restate-server >= 1.4,
            otherwise the service discovery will fail.
        ingress_private (bool, optional): When set to True this service, with all its handlers, cannot be invoked from the restate-server
            HTTP and Kafka ingress, but only from other services.
            NOTE: You can set this field only if you register this service against restate-server >= 1.4,
            otherwise the service discovery will fail.
        invocation_retry_policy (InvocationRetryPolicy, optional): Retry policy applied for all invocations to this workflow.
        invocation_context_managers (List[Callable[[], AsyncContextManager[None]]], optional): List of context manager factories that will be applied to each invocation of this workflow.
    """

    handlers: typing.Dict[str, Handler[typing.Any, typing.Any]]

    def __init__(
        self,
        name,
        description: typing.Optional[str] = None,
        metadata: typing.Optional[typing.Dict[str, str]] = None,
        inactivity_timeout: typing.Optional[timedelta] = None,
        abort_timeout: typing.Optional[timedelta] = None,
        journal_retention: typing.Optional[timedelta] = None,
        idempotency_retention: typing.Optional[timedelta] = None,
        enable_lazy_state: typing.Optional[bool] = None,
        ingress_private: typing.Optional[bool] = None,
        invocation_retry_policy: typing.Optional[InvocationRetryPolicy] = None,
        invocation_context_managers: typing.Optional[
            typing.List[typing.Callable[[], typing.AsyncContextManager[None]]]
        ] = None,
    ):
        self.service_tag = ServiceTag("workflow", name, description, metadata)
        self.handlers = {}
        self.inactivity_timeout = inactivity_timeout
        self.abort_timeout = abort_timeout
        self.journal_retention = journal_retention
        self.idempotency_retention = idempotency_retention
        self.enable_lazy_state = enable_lazy_state
        self.ingress_private = ingress_private
        self.invocation_retry_policy = invocation_retry_policy
        self.context_managers = invocation_context_managers

    @property
    def name(self):
        """
        Returns the name of the object.
        """
        return self.service_tag.name

    def main(
        self,
        name: typing.Optional[str] = None,
        accept: str = "application/json",
        content_type: str = "application/json",
        input_serde: Serde[I] = DefaultSerde[I](),  # type: ignore
        output_serde: Serde[O] = DefaultSerde[O](),  # type: ignore
        metadata: typing.Optional[typing.Dict[str, str]] = None,
        inactivity_timeout: typing.Optional[timedelta] = None,
        abort_timeout: typing.Optional[timedelta] = None,
        journal_retention: typing.Optional[timedelta] = None,
        workflow_retention: typing.Optional[timedelta] = None,
        enable_lazy_state: typing.Optional[bool] = None,
        ingress_private: typing.Optional[bool] = None,
        invocation_retry_policy: typing.Optional[InvocationRetryPolicy] = None,
        invocation_context_managers: typing.Optional[
            typing.List[typing.Callable[[], typing.AsyncContextManager[None]]]
        ] = None,
    ) -> typing.Callable[[T], T]:
        """
        Mark this handler as a workflow entry point.

        Args:
            name: The name of the handler.
            accept: Set the acceptable content type when ingesting HTTP requests. Wildcards can be used, e.g.
                `application/*` or `*/*`. Default "application/json".
            content_type: The content type of the request. Default "application/json".
            input_serde: The serializer/deserializer for the input parameter.
            output_serde: The serializer/deserializer for the output result.
            metadata: Handler metadata, as propagated in the Admin REST API.
            inactivity_timeout: This timer guards against stalled invocations. Once it expires, Restate triggers a graceful
                termination by asking the invocation to suspend (which preserves intermediate progress).
                The abort_timeout is used to abort the invocation, in case it doesn't react to the request to suspend.
                This overrides the inactivity timeout set for the service and the default set in restate-server.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            abort_timeout: This timer guards against stalled invocations that are supposed to terminate. The abort timeout
                is started after the inactivity_timeout has expired and the invocation has been asked to gracefully terminate.
                Once the timer expires, it will abort the invocation.
                This timer potentially *interrupts* user code. If the user code needs longer to gracefully terminate, then this value needs to be set accordingly.
                This overrides the abort timeout set for the service and the default set in restate-server.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            journal_retention: The journal retention for invocations to this handler.
                In case the request has an idempotency key, the idempotency_retention caps the journal retention time.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            workflow_retention: The retention duration for this workflow handler.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            enable_lazy_state: When set to True, lazy state will be enabled for all invocations to this handler.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            ingress_private: When set to True this handler cannot be invoked from the restate-server HTTP and Kafka ingress,
                but only from other services.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            invocation_retry_policy (InvocationRetryPolicy, optional): Retry policy applied for all invocations to this workflow.
        """
        return self._add_handler(
            name,
            kind="workflow",
            accept=accept,
            content_type=content_type,
            input_serde=input_serde,
            output_serde=output_serde,
            metadata=metadata,
            inactivity_timeout=inactivity_timeout,
            abort_timeout=abort_timeout,
            journal_retention=journal_retention,
            idempotency_retention=None,
            workflow_retention=workflow_retention,
            enable_lazy_state=enable_lazy_state,
            ingress_private=ingress_private,
            invocation_retry_policy=invocation_retry_policy,
            invocation_context_managers=invocation_context_managers,
        )

    def handler(
        self,
        name: typing.Optional[str] = None,
        accept: str = "application/json",
        content_type: str = "application/json",
        input_serde: Serde[I] = DefaultSerde[I](),  # type: ignore
        output_serde: Serde[O] = DefaultSerde[O](),  # type: ignore
        metadata: typing.Optional[typing.Dict[str, str]] = None,
        inactivity_timeout: typing.Optional[timedelta] = None,
        abort_timeout: typing.Optional[timedelta] = None,
        journal_retention: typing.Optional[timedelta] = None,
        idempotency_retention: typing.Optional[timedelta] = None,
        enable_lazy_state: typing.Optional[bool] = None,
        ingress_private: typing.Optional[bool] = None,
        invocation_retry_policy: typing.Optional[InvocationRetryPolicy] = None,
        invocation_context_managers: typing.Optional[
            typing.List[typing.Callable[[], typing.AsyncContextManager[None]]]
        ] = None,
    ) -> typing.Callable[[T], T]:
        """
        Decorator for defining a handler function.

        Args:
            name: The name of the handler.
            accept: Set the acceptable content type when ingesting HTTP requests. Wildcards can be used, e.g.
                `application/*` or `*/*`. Default "application/json".
            content_type: The content type of the request. Default "application/json".
            input_serde: The serializer/deserializer for the input parameter.
            output_serde: The serializer/deserializer for the output result.
            metadata: Handler metadata, as propagated in the Admin REST API.
            inactivity_timeout: This timer guards against stalled invocations. Once it expires, Restate triggers a graceful
                termination by asking the invocation to suspend (which preserves intermediate progress).
                The abort_timeout is used to abort the invocation, in case it doesn't react to the request to suspend.
                This overrides the inactivity timeout set for the service and the default set in restate-server.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            abort_timeout: This timer guards against stalled invocations that are supposed to terminate. The abort timeout
                is started after the inactivity_timeout has expired and the invocation has been asked to gracefully terminate.
                Once the timer expires, it will abort the invocation.
                This timer potentially *interrupts* user code. If the user code needs longer to gracefully terminate, then this value needs to be set accordingly.
                This overrides the abort timeout set for the service and the default set in restate-server.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            journal_retention: The journal retention for invocations to this handler.
                In case the request has an idempotency key, the idempotency_retention caps the journal retention time.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            idempotency_retention: The retention duration of idempotent requests to this service.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            enable_lazy_state: When set to True, lazy state will be enabled for all invocations to this handler.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            ingress_private: When set to True this handler cannot be invoked from the restate-server HTTP and Kafka ingress,
                but only from other services.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            invocation_retry_policy (InvocationRetryPolicy, optional): Retry policy applied for all invocations to this handler.
            invocation_context_managers (List[Callable[[], AsyncContextManager[None]]], optional): List of context manager factories that will be applied to each invocation of this handler.
        """
        return self._add_handler(
            name,
            "shared",
            accept,
            content_type,
            input_serde,
            output_serde,
            metadata,
            inactivity_timeout,
            abort_timeout,
            journal_retention,
            idempotency_retention,
            None,
            enable_lazy_state,
            ingress_private,
            invocation_retry_policy,
            invocation_context_managers,
        )

    # pylint: disable=R0914
    def _add_handler(
        self,
        name: typing.Optional[str] = None,
        kind: typing.Literal["workflow", "shared", "exclusive"] = "shared",
        accept: str = "application/json",
        content_type: str = "application/json",
        input_serde: Serde[I] = DefaultSerde[I](),  # type: ignore
        output_serde: Serde[O] = DefaultSerde[O](),  # type: ignore
        metadata: typing.Optional[typing.Dict[str, str]] = None,
        inactivity_timeout: typing.Optional[timedelta] = None,
        abort_timeout: typing.Optional[timedelta] = None,
        journal_retention: typing.Optional[timedelta] = None,
        idempotency_retention: typing.Optional[timedelta] = None,
        workflow_retention: typing.Optional[timedelta] = None,
        enable_lazy_state: typing.Optional[bool] = None,
        ingress_private: typing.Optional[bool] = None,
        invocation_retry_policy: typing.Optional["InvocationRetryPolicy"] = None,
        invocation_context_managers: typing.Optional[
            typing.List[typing.Callable[[], typing.AsyncContextManager[None]]]
        ] = None,
    ) -> typing.Callable[[T], T]:
        """
        Decorator for defining a handler function.

        Args:
            name: The name of the handler.
            kind: The kind of handler (workflow, shared, exclusive). Default "shared".
            accept: Set the acceptable content type when ingesting HTTP requests. Wildcards can be used, e.g.
                `application/*` or `*/*`. Default "application/json".
            content_type: The content type of the request. Default "application/json".
            input_serde: The serializer/deserializer for the input parameter.
            output_serde: The serializer/deserializer for the output result.
            metadata: Handler metadata, as propagated in the Admin REST API.
            inactivity_timeout: This timer guards against stalled invocations. Once it expires, Restate triggers a graceful
                termination by asking the invocation to suspend (which preserves intermediate progress).
                The abort_timeout is used to abort the invocation, in case it doesn't react to the request to suspend.
                This overrides the inactivity timeout set for the service and the default set in restate-server.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            abort_timeout: This timer guards against stalled invocations that are supposed to terminate. The abort timeout
                is started after the inactivity_timeout has expired and the invocation has been asked to gracefully terminate.
                Once the timer expires, it will abort the invocation.
                This timer potentially *interrupts* user code. If the user code needs longer to gracefully terminate, then this value needs to be set accordingly.
                This overrides the abort timeout set for the service and the default set in restate-server.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            journal_retention: The journal retention for invocations to this handler.
                In case the request has an idempotency key, the idempotency_retention caps the journal retention time.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            idempotency_retention: The retention duration of idempotent requests to this service.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            workflow_retention: The retention duration for this workflow handler.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            enable_lazy_state: When set to True, lazy state will be enabled for all invocations to this handler.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            ingress_private: When set to True this handler cannot be invoked from the restate-server HTTP and Kafka ingress,
                but only from other services.
                NOTE: You can set this field only if you register this service against restate-server >= 1.4,
                otherwise the service discovery will fail.
            invocation_retry_policy (InvocationRetryPolicy, optional): Retry policy applied for all invocations to this handler.

        Returns:
            Callable: The decorated function.

        Raises:
            ValueError: If the handler name is not provided.

        Example:
            @service.handler()
            def my_handler_func(ctx, request):
                # handler logic
                pass
        """
        handler_io = HandlerIO[I, O](accept, content_type, input_serde, output_serde)

        def wrapper(fn):
            @wraps(fn)
            def wrapped(*args, **kwargs):
                return fn(*args, **kwargs)

            signature = inspect.signature(fn, eval_str=True)
            description = inspect.getdoc(fn)
            combined_context_managers = (
                (self.context_managers or []) + (invocation_context_managers or [])
                if self.context_managers or invocation_context_managers
                else None
            )
            handler = make_handler(
                service_tag=self.service_tag,
                handler_io=handler_io,
                name=name,
                kind=kind,
                wrapped=wrapped,
                signature=signature,
                description=description,
                metadata=metadata,
                inactivity_timeout=inactivity_timeout,
                abort_timeout=abort_timeout,
                journal_retention=journal_retention,
                idempotency_retention=idempotency_retention,
                workflow_retention=workflow_retention,
                enable_lazy_state=enable_lazy_state,
                ingress_private=ingress_private,
                invocation_retry_policy=invocation_retry_policy,
                context_managers=combined_context_managers,
            )
            self.handlers[handler.name] = handler
            return wrapped

        return typing.cast(typing.Callable[[T], T], wrapper)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""greeter.py"""
# pylint: disable=C0116
# pylint: disable=W0613
# pylint: disable=C0115
# pylint: disable=R0903
# pylint: disable=C0301

import typing

from pydantic import BaseModel
from restate import wait_completed, Service, Context


# models
class GreetingRequest(BaseModel):
    name: str


class Greeting(BaseModel):
    messages: typing.List[str]


class Message(BaseModel):
    role: str
    content: str


concurrent_greeter = Service("concurrent_greeter")


@concurrent_greeter.handler()
async def greet(ctx: Context, req: GreetingRequest) -> Greeting:
    claude = ctx.service_call(claude_sonnet, arg=Message(role="user", content=f"please greet {req.name}"))
    openai = ctx.service_call(open_ai, arg=Message(role="user", content=f"please greet {req.name}"))

    pending, done = await wait_completed(claude, openai)

    # collect the completed greetings
    greetings = [await f for f in done]

    # cancel the pending calls
    for f in pending:
        await f.cancel_invocation()  # type: ignore

    return Greeting(messages=greetings)


# not really interesting, just for this demo:


@concurrent_greeter.handler()
async def claude_sonnet(ctx: Context, req: Message) -> str:
    return f"Bonjour {req.content[13:]}!"


@concurrent_greeter.handler()
async def open_ai(ctx: Context, req: Message) -> str:
    return f"Hello {req.content[13:]}!"



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
Type definitions for the Restate client.
"""

import abc
from datetime import timedelta
import typing


from .context import HandlerType

I = typing.TypeVar("I")
O = typing.TypeVar("O")


class RestateClientSendHandle:
    """
    A handle for a send operation.
    This is used to track the status of a send operation.
    """

    def __init__(self, invocation_id: str, status_code: int):
        self.invocation_id = invocation_id
        self.status_code = status_code


class HttpError(Exception):
    """
    An error that occurs during an HTTP request.
    """

    def __init__(self, status_code: int, message: str, body: str | None = None):
        super().__init__(f"HTTP {status_code}: {message}")
        self.status_code = status_code
        self.message = message
        self.body = body


class RestateClient(abc.ABC):
    """
    An abstract base class for a Restate client.
    This class defines the interface for a Restate client.
    """

    @abc.abstractmethod
    async def service_call(
        self,
        tpe: HandlerType[I, O],
        arg: I,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> O:
        """Make an RPC call to the given handler"""
        pass

    @abc.abstractmethod
    async def service_send(
        self,
        tpe: HandlerType[I, O],
        arg: I,
        send_delay: typing.Optional[timedelta] = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateClientSendHandle:
        """Make a send operation to the given handler"""
        pass

    @abc.abstractmethod
    async def object_call(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> O:
        """Make an RPC call to the given object handler"""
        pass

    @abc.abstractmethod
    async def object_send(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        send_delay: typing.Optional[timedelta] = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateClientSendHandle:
        """Make a send operation to the given object handler"""
        pass

    @abc.abstractmethod
    async def workflow_call(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> O:
        """Make an RPC call to the given workflow handler"""
        pass

    @abc.abstractmethod
    async def workflow_send(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        send_delay: typing.Optional[timedelta] = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateClientSendHandle:
        """Make a send operation to the given workflow handler"""
        pass

    @abc.abstractmethod
    async def generic_call(
        self,
        service: str,
        handler: str,
        arg: bytes,
        key: str | None = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> bytes:
        """Make a generic RPC call to the given service and handler"""
        pass

    @abc.abstractmethod
    async def generic_send(
        self,
        service: str,
        handler: str,
        arg: bytes,
        key: str | None = None,
        send_delay: timedelta | None = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateClientSendHandle:
        """Make a generic send operation to the given service and handler"""
        pass



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
This module contains the ASGI types definitions.

:see: https://github.com/django/asgiref/blob/main/asgiref/typing.py
"""

import asyncio
from typing import Awaitable, Callable, Dict, Iterable, List, Tuple, Union, TypedDict, Literal, Optional, Any


class ASGIVersions(TypedDict):
    """ASGI Versions"""

    spec_version: str
    version: Union[Literal["2.0"], Literal["3.0"]]


class Scope(TypedDict):
    """ASGI Scope"""

    type: Literal["http"]
    asgi: ASGIVersions
    http_version: str
    method: str
    scheme: str
    path: str
    raw_path: bytes
    query_string: bytes
    root_path: str
    headers: Iterable[Tuple[bytes, bytes]]
    client: Optional[Tuple[str, int]]
    server: Optional[Tuple[str, Optional[int]]]
    extensions: Optional[Dict[str, Dict[object, object]]]


class RestateEvent(TypedDict):
    """An event that represents a run completion"""

    type: Literal["restate.run_completed"]
    data: Optional[Dict[str, Any]]


class HTTPRequestEvent(TypedDict):
    """ASGI Request event"""

    type: Literal["http.request"]
    body: bytes
    more_body: bool


class HTTPResponseStartEvent(TypedDict):
    """ASGI Response start event"""

    type: Literal["http.response.start"]
    status: int
    headers: Iterable[Tuple[bytes, bytes]]
    trailers: bool


class HTTPResponseBodyEvent(TypedDict):
    """ASGI Response body event"""

    type: Literal["http.response.body"]
    body: bytes
    more_body: bool


ASGIReceiveEvent = HTTPRequestEvent


ASGISendEvent = Union[HTTPResponseStartEvent, HTTPResponseBodyEvent]

Receive = Callable[[], Awaitable[ASGIReceiveEvent]]
Send = Callable[[ASGISendEvent], Awaitable[None]]

ASGIApp = Callable[
    [
        Scope,
        Receive,
        Send,
    ],
    Awaitable[None],
]


class RestateLambdaRequest(TypedDict):
    """
    Restate Lambda request

    :see: https://github.com/restatedev/restate/blob/1a10c05b16b387191060b49faffb0335ee97e96d/crates/service-client/src/lambda.rs#L297 # pylint: disable=line-too-long
    """

    path: str
    httpMethod: str
    headers: Dict[str, str]
    body: str
    isBase64Encoded: bool


class RestateLambdaResponse(TypedDict):
    """
    Restate Lambda response

    :see: https://github.com/restatedev/restate/blob/1a10c05b16b387191060b49faffb0335ee97e96d/crates/service-client/src/lambda.rs#L310 # pylint: disable=line-too-long
    """

    statusCode: int
    headers: Dict[str, str]
    body: str
    isBase64Encoded: bool


RestateLambdaHandler = Callable[[RestateLambdaRequest, Any], RestateLambdaResponse]

RestateAppT = Any  # Union[ASGIApp, RestateLambdaHandler]


def header_to_binary(headers: Iterable[Tuple[str, str]]) -> List[Tuple[bytes, bytes]]:
    """Convert a list of headers to a list of binary headers."""
    return [(k.encode("utf-8"), v.encode("utf-8")) for k, v in headers]


def binary_to_header(headers: Iterable[Tuple[bytes, bytes]]) -> List[Tuple[str, str]]:
    """Convert a list of binary headers to a list of headers."""
    return [(k.decode("utf-8"), v.decode("utf-8")) for k, v in headers]


class ReceiveChannel:
    """ASGI receive channel."""

    def __init__(self, receive: Receive) -> None:
        self._queue = asyncio.Queue[Union[ASGIReceiveEvent, RestateEvent]]()
        self._http_input_closed = asyncio.Event()
        self._disconnected = asyncio.Event()

        async def loop():
            """Receive loop."""
            while not self._disconnected.is_set():
                event = await receive()
                if event.get("type") == "http.request" and not event.get("more_body", False):
                    self._http_input_closed.set()
                elif event.get("type") == "http.disconnect":
                    self._http_input_closed.set()
                    self._disconnected.set()
                await self._queue.put(event)

        self._task = asyncio.create_task(loop())

    async def __call__(self) -> ASGIReceiveEvent | RestateEvent:
        """Get the next message."""
        what = await self._queue.get()
        self._queue.task_done()
        return what

    async def block_until_http_input_closed(self) -> None:
        """Wait until the HTTP input is closed"""
        await self._http_input_closed.wait()

    async def enqueue_restate_event(self, what: RestateEvent):
        """Add a message."""
        await self._queue.put(what)

    async def close(self):
        """Close the channel."""
        self._http_input_closed.set()
        self._disconnected.set()
        if self._task.done():
            return
        self._task.cancel()
        try:
            await self._task
        except asyncio.CancelledError:
            pass



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
This is a basic remote client for the Restate service.
"""

from datetime import timedelta
import httpx
import typing
from contextlib import asynccontextmanager

from .client_types import RestateClient, RestateClientSendHandle, HttpError

from .context import HandlerType
from .serde import BytesSerde, JsonSerde, Serde
from .handler import handler_from_callable

I = typing.TypeVar("I")
O = typing.TypeVar("O")


class Client(RestateClient):
    """
    A basic client for connecting to the Restate service.
    """

    def __init__(self, client: httpx.AsyncClient, headers: typing.Optional[dict] = None):
        self.headers = headers or {}
        self.client = client

    async def do_call(
        self,
        tpe: HandlerType[I, O],
        parameter: I,
        key: typing.Optional[str] = None,
        send_delay: typing.Optional[timedelta] = None,
        send: bool = False,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
        force_json_output: bool = False,
    ) -> O:
        """Make an RPC call to the given handler"""
        target_handler = handler_from_callable(tpe)
        input_serde = target_handler.handler_io.input_serde
        input_type = target_handler.handler_io.input_type

        if input_type is not None and input_type.is_void:
            content_type = None
        else:
            content_type = target_handler.handler_io.content_type

        if headers is None:
            headers = {}
        if headers.get("Content-Type") is None and content_type is not None:
            headers["Content-Type"] = content_type

        service = target_handler.service_tag.name
        handler = target_handler.name
        output_serde = target_handler.handler_io.output_serde if force_json_output is False else JsonSerde()

        return await self.do_raw_call(
            service=service,
            handler=handler,
            input_param=parameter,
            input_serde=input_serde,
            output_serde=output_serde,
            key=key,
            send_delay=send_delay,
            send=send,
            idempotency_key=idempotency_key,
            headers=headers,
        )

    async def do_raw_call(
        self,
        service: str,
        handler: str,
        input_param: I,
        input_serde: Serde[I],
        output_serde: Serde[O],
        key: typing.Optional[str] = None,
        send_delay: typing.Optional[timedelta] = None,
        send: bool = False,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> O:
        """Make an RPC call to the given handler"""
        parameter = input_serde.serialize(input_param)
        if headers is not None:
            headers_kvs = list(headers.items())
        else:
            headers_kvs = []
        if send_delay is not None:
            ms = int(send_delay.total_seconds() * 1000)
        else:
            ms = None

        res = await self.post(
            service=service,
            handler=handler,
            send=send,
            content=parameter,
            headers=headers_kvs,
            key=key,
            delay=ms,
            idempotency_key=idempotency_key,
        )
        return output_serde.deserialize(res)  # type: ignore

    async def post(
        self,
        /,
        service: str,
        handler: str,
        send: bool,
        content: bytes,
        headers: typing.List[typing.Tuple[(str, str)]] | None = None,
        key: str | None = None,
        delay: int | None = None,
        idempotency_key: str | None = None,
    ) -> bytes:
        """
        Send a POST request to the Restate service.
        """
        endpoint = service
        if key:
            endpoint += f"/{key}"
        endpoint += f"/{handler}"
        if send:
            endpoint += "/send"
            if delay is not None:
                endpoint = endpoint + f"?delay={delay}"
        dict_headers = dict(headers) if headers is not None else {}
        if idempotency_key is not None:
            dict_headers["Idempotency-Key"] = idempotency_key
        res = await self.client.post(endpoint, headers=dict_headers, content=content)
        if res.status_code >= 400:
            raise HttpError(res.status_code, res.reason_phrase, res.text)
        return res.content

    @typing.final
    async def service_call(
        self,
        tpe: HandlerType[I, O],
        arg: I,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> O:
        coro = await self.do_call(tpe, arg, idempotency_key=idempotency_key, headers=headers)
        return coro

    async def service_send(
        self,
        tpe: HandlerType[I, O],
        arg: I,
        send_delay: typing.Optional[timedelta] = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateClientSendHandle:
        send_handle = await self.do_call(
            tpe,
            parameter=arg,
            send=True,
            send_delay=send_delay,
            idempotency_key=idempotency_key,
            headers=headers,
            force_json_output=True,
        )

        send = typing.cast(typing.Dict[str, str], send_handle)

        return RestateClientSendHandle(send.get("invocationId", ""), 200)  # TODO: verify

    async def object_call(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> O:
        coro = await self.do_call(tpe, arg, key, idempotency_key=idempotency_key, headers=headers)
        return coro

    async def object_send(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        send_delay: typing.Optional[timedelta] = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateClientSendHandle:
        send_handle = await self.do_call(
            tpe,
            parameter=arg,
            key=key,
            send=True,
            send_delay=send_delay,
            idempotency_key=idempotency_key,
            headers=headers,
            force_json_output=True,
        )

        send = typing.cast(typing.Dict[str, str], send_handle)

        return RestateClientSendHandle(send.get("invocationId", ""), 200)  # TODO: verify

    async def workflow_call(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> O:
        return await self.object_call(tpe, key, arg, idempotency_key=idempotency_key, headers=headers)

    async def workflow_send(
        self,
        tpe: HandlerType[I, O],
        key: str,
        arg: I,
        send_delay: typing.Optional[timedelta] = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateClientSendHandle:
        return await self.object_send(
            tpe,
            key,
            arg,
            send_delay=send_delay,
            idempotency_key=idempotency_key,
            headers=headers,
        )

    async def generic_call(
        self,
        service: str,
        handler: str,
        arg: bytes,
        key: str | None = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> bytes:
        serde = BytesSerde()
        call_handle = await self.do_raw_call(
            service=service,
            handler=handler,
            input_param=arg,
            input_serde=serde,
            output_serde=serde,
            key=key,
            idempotency_key=idempotency_key,
            headers=headers,
        )
        return call_handle

    async def generic_send(
        self,
        service: str,
        handler: str,
        arg: bytes,
        key: str | None = None,
        send_delay: timedelta | None = None,
        idempotency_key: str | None = None,
        headers: typing.Dict[str, str] | None = None,
    ) -> RestateClientSendHandle:
        serde = BytesSerde()
        output_serde: Serde[dict] = JsonSerde()

        send_handle_json = await self.do_raw_call(
            service=service,
            handler=handler,
            input_param=arg,
            input_serde=serde,
            output_serde=output_serde,
            key=key,
            send_delay=send_delay,
            send=True,
            idempotency_key=idempotency_key,
            headers=headers,
        )

        return RestateClientSendHandle(send_handle_json.get("invocationId", ""), 200)  # TODO: verify


@asynccontextmanager
async def create_client(
    ingress: str, headers: typing.Optional[dict] = None
) -> typing.AsyncGenerator[RestateClient, None]:
    """
    Create a new Restate client.
    """
    async with httpx.AsyncClient(base_url=ingress, headers=headers, http2=True) as http_client:
        yield Client(http_client, headers)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
This module contains the optional OpenAI integration for Restate.
"""

from .runner_wrapper import Runner, DurableModelCalls, continue_on_terminal_errors, raise_terminal_errors

__all__ = [
    "DurableModelCalls",
    "continue_on_terminal_errors",
    "raise_terminal_errors",
    "Runner",
]



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
import restate

from virtual_object import increment, count, counter

#
# uv run examples/harness.py
#


async def main():
    app = restate.app([counter])
    async with restate.create_test_harness(app) as harness:
        await harness.client.object_call(increment, key="a", arg=5)
        await harness.client.object_call(increment, key="a", arg=5)

        current_count = await harness.client.object_call(count, key="a", arg=None)

        print(f"Current count for 'a': {current_count}")

        send = await harness.client.object_send(increment, key="b", arg=-10, idempotency_key="op1")

        print(f"Sent increment to 'b', invocation ID: {send.invocation_id}")


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
Holds the discovery API objects as defined by the restate protocol.
Note that the classes defined here do not use snake case, because they
are intended to be serialized to JSON, and their cases must remain in the
case that the restate server understands.
"""

# disable to few parameters
# pylint: disable=R0903
# pylint: disable=C0301
# pylint: disable=C0115
# pylint: disable=C0103
# pylint: disable=W0622
# pylint: disable=R0913,
# pylint: disable=R0917,

import json
import typing
from enum import Enum
from typing import Dict, Optional, Any, List, get_args, get_origin


from restate.endpoint import Endpoint as RestateEndpoint
from restate.handler import TypeHint
from restate.object import VirtualObject
from restate.workflow import Workflow


class ProtocolMode(Enum):
    BIDI_STREAM = "BIDI_STREAM"
    REQUEST_RESPONSE = "REQUEST_RESPONSE"


class ServiceType(Enum):
    VIRTUAL_OBJECT = "VIRTUAL_OBJECT"
    SERVICE = "SERVICE"
    WORKFLOW = "WORKFLOW"


class ServiceHandlerType(Enum):
    WORKFLOW = "WORKFLOW"
    EXCLUSIVE = "EXCLUSIVE"
    SHARED = "SHARED"


class InputPayload:
    def __init__(self, required: bool, contentType: str, jsonSchema: Optional[Any] = None):
        self.required = required
        self.contentType = contentType
        self.jsonSchema = jsonSchema


class OutputPayload:
    def __init__(
        self, setContentTypeIfEmpty: bool, contentType: Optional[str] = None, jsonSchema: Optional[Any] = None
    ):
        self.contentType = contentType
        self.setContentTypeIfEmpty = setContentTypeIfEmpty
        self.jsonSchema = jsonSchema


class Handler:
    # pylint: disable=R0902,R0914
    def __init__(
        self,
        name: str,
        ty: Optional[ServiceHandlerType] = None,
        input: Optional[InputPayload | Dict[str, str]] = None,
        output: Optional[OutputPayload] = None,
        description: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None,
        inactivityTimeout: Optional[int] = None,
        abortTimeout: Optional[int] = None,
        journalRetention: Optional[int] = None,
        idempotencyRetention: Optional[int] = None,
        workflowCompletionRetention: Optional[int] = None,
        enableLazyState: Optional[bool] = None,
        ingressPrivate: Optional[bool] = None,
        retryPolicyInitialInterval: Optional[int] = None,
        retryPolicyMaxInterval: Optional[int] = None,
        retryPolicyMaxAttempts: Optional[int] = None,
        retryPolicyExponentiationFactor: Optional[float] = None,
        retryPolicyOnMaxAttempts: Optional[str] = None,
    ):
        self.name = name
        self.ty = ty
        self.input = input
        self.output = output
        self.documentation = description
        self.metadata = metadata
        self.inactivityTimeout = inactivityTimeout
        self.abortTimeout = abortTimeout
        self.journalRetention = journalRetention
        self.idempotencyRetention = idempotencyRetention
        self.workflowCompletionRetention = workflowCompletionRetention
        self.enableLazyState = enableLazyState
        self.ingressPrivate = ingressPrivate
        self.retryPolicyInitialInterval = retryPolicyInitialInterval
        self.retryPolicyMaxInterval = retryPolicyMaxInterval
        self.retryPolicyMaxAttempts = retryPolicyMaxAttempts
        self.retryPolicyExponentiationFactor = retryPolicyExponentiationFactor
        self.retryPolicyOnMaxAttempts = retryPolicyOnMaxAttempts


class Service:
    # pylint: disable=R0902,R0914
    def __init__(
        self,
        name: str,
        ty: ServiceType,
        handlers: List[Handler],
        description: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None,
        inactivityTimeout: Optional[int] = None,
        abortTimeout: Optional[int] = None,
        journalRetention: Optional[int] = None,
        idempotencyRetention: Optional[int] = None,
        enableLazyState: Optional[bool] = None,
        ingressPrivate: Optional[bool] = None,
        retryPolicyInitialInterval: Optional[int] = None,
        retryPolicyMaxInterval: Optional[int] = None,
        retryPolicyMaxAttempts: Optional[int] = None,
        retryPolicyExponentiationFactor: Optional[float] = None,
        retryPolicyOnMaxAttempts: Optional[str] = None,
    ):
        self.name = name
        self.ty = ty
        self.handlers = handlers
        self.documentation = description
        self.metadata = metadata
        self.inactivityTimeout = inactivityTimeout
        self.abortTimeout = abortTimeout
        self.journalRetention = journalRetention
        self.idempotencyRetention = idempotencyRetention
        self.enableLazyState = enableLazyState
        self.ingressPrivate = ingressPrivate
        self.retryPolicyInitialInterval = retryPolicyInitialInterval
        self.retryPolicyMaxInterval = retryPolicyMaxInterval
        self.retryPolicyMaxAttempts = retryPolicyMaxAttempts
        self.retryPolicyExponentiationFactor = retryPolicyExponentiationFactor
        self.retryPolicyOnMaxAttempts = retryPolicyOnMaxAttempts


class Endpoint:
    def __init__(
        self, protocolMode: ProtocolMode, minProtocolVersion: int, maxProtocolVersion: int, services: List[Service]
    ):
        self.protocolMode = protocolMode
        self.minProtocolVersion = minProtocolVersion
        self.maxProtocolVersion = maxProtocolVersion
        self.services = services


PROTOCOL_MODES = {"bidi": ProtocolMode.BIDI_STREAM, "request_response": ProtocolMode.REQUEST_RESPONSE}

SERVICE_TYPES = {"service": ServiceType.SERVICE, "object": ServiceType.VIRTUAL_OBJECT, "workflow": ServiceType.WORKFLOW}

HANDLER_TYPES = {
    "exclusive": ServiceHandlerType.EXCLUSIVE,
    "shared": ServiceHandlerType.SHARED,
    "workflow": ServiceHandlerType.WORKFLOW,
}


class PythonClassEncoder(json.JSONEncoder):
    """
    Serialize Python objects as JSON
    """

    def default(self, o):
        if isinstance(o, Enum):
            return o.value
        return {key: value for key, value in o.__dict__.items() if value is not None}


# pylint: disable=R0911
def type_hint_to_json_schema(type_hint: Any) -> Any:
    """
    Convert a Python type hint to a JSON schema.

    """
    origin = get_origin(type_hint) or type_hint
    args = get_args(type_hint)
    if origin is str:
        return {"type": "string"}
    if origin is int:
        return {"type": "integer"}
    if origin is float:
        return {"type": "number"}
    if origin is bool:
        return {"type": "boolean"}
    if origin is list:
        items = type_hint_to_json_schema(args[0] if args else Any)
        return {"type": "array", "items": items}
    if origin is dict:
        return {"type": "object"}
    if origin is None:
        return {"type": "null"}
    # Default to all valid schema
    return True


def json_schema_from_type_hint(type_hint: Optional[TypeHint[Any]]) -> Any:
    """
    Convert a type hint to a JSON schema.
    """
    if not type_hint:
        return None
    if not type_hint.annotation:
        return None
    if type_hint.generate_json_schema is not None:
        return type_hint.generate_json_schema()
    return type_hint_to_json_schema(type_hint.annotation)


# pylint: disable=R0912,R0915
def compute_discovery_json(
    endpoint: RestateEndpoint, version: int, discovered_as: typing.Literal["bidi", "request_response"]
) -> str:
    """
    return restate's discovery object as JSON
    """

    ep = compute_discovery(endpoint, discovered_as)

    # Validate that new discovery fields aren't used with older protocol versions
    if version <= 3:
        for service in ep.services:
            if service.retryPolicyInitialInterval is not None:
                raise ValueError("retryPolicyInitialInterval is only supported in discovery protocol version 4")
            if service.retryPolicyMaxInterval is not None:
                raise ValueError("retryPolicyMaxInterval is only supported in discovery protocol version 4")
            if service.retryPolicyMaxAttempts is not None:
                raise ValueError("retryPolicyMaxAttempts is only supported in discovery protocol version 4")
            if service.retryPolicyExponentiationFactor is not None:
                raise ValueError("retryPolicyExponentiationFactor is only supported in discovery protocol version 4")
            if service.retryPolicyOnMaxAttempts is not None:
                raise ValueError("retryPolicyOnMaxAttempts is only supported in discovery protocol version 4")

            for handler in service.handlers:
                if handler.retryPolicyInitialInterval is not None:
                    raise ValueError("retryPolicyInitialInterval is only supported in discovery protocol version 4")
                if handler.retryPolicyMaxInterval is not None:
                    raise ValueError("retryPolicyMaxInterval is only supported in discovery protocol version 4")
                if handler.retryPolicyMaxAttempts is not None:
                    raise ValueError("retryPolicyMaxAttempts is only supported in discovery protocol version 4")
                if handler.retryPolicyExponentiationFactor is not None:
                    raise ValueError(
                        "retryPolicyExponentiationFactor is only supported in discovery protocol version 4"
                    )
                if handler.retryPolicyOnMaxAttempts is not None:
                    raise ValueError("retryPolicyOnMaxAttempts is only supported in discovery protocol version 4")

    if version <= 2:
        # Check for new discovery fields in version 3 that shouldn't be used in version 2 or lower
        for service in ep.services:
            if service.inactivityTimeout is not None:
                raise ValueError("inactivityTimeout is only supported in discovery protocol version 3")
            if service.abortTimeout is not None:
                raise ValueError("abortTimeout is only supported in discovery protocol version 3")
            if service.idempotencyRetention is not None:
                raise ValueError("idempotencyRetention is only supported in discovery protocol version 3")
            if service.journalRetention is not None:
                raise ValueError("journalRetention is only supported in discovery protocol version 3")
            if service.enableLazyState is not None:
                raise ValueError("enableLazyState is only supported in discovery protocol version 3")
            if service.ingressPrivate is not None:
                raise ValueError("ingressPrivate is only supported in discovery protocol version 3")

            for handler in service.handlers:
                if handler.inactivityTimeout is not None:
                    raise ValueError("inactivityTimeout is only supported in discovery protocol version 3")
                if handler.abortTimeout is not None:
                    raise ValueError("abortTimeout is only supported in discovery protocol version 3")
                if handler.idempotencyRetention is not None:
                    raise ValueError("idempotencyRetention is only supported in discovery protocol version 3")
                if handler.journalRetention is not None:
                    raise ValueError("journalRetention is only supported in discovery protocol version 3")
                if handler.workflowCompletionRetention is not None:
                    raise ValueError("workflowCompletionRetention is only supported in discovery protocol version 3")
                if handler.enableLazyState is not None:
                    raise ValueError("enableLazyState is only supported in discovery protocol version 3")
                if handler.ingressPrivate is not None:
                    raise ValueError("ingressPrivate is only supported in discovery protocol version 3")

    json_str = json.dumps(ep, cls=PythonClassEncoder, allow_nan=False)
    return json_str


def compute_discovery(endpoint: RestateEndpoint, discovered_as: typing.Literal["bidi", "request_response"]) -> Endpoint:
    """
    return restate's discovery object for an endpoint
    """
    services: typing.List[Service] = []

    for service in endpoint.services.values():
        service_type = SERVICE_TYPES[service.service_tag.kind]
        service_handlers = []
        for handler in service.handlers.values():
            # type
            if handler.kind:
                ty = HANDLER_TYPES[handler.kind]
            else:
                ty = None
            # input
            inp: Optional[InputPayload | Dict[str, str]] = None
            if handler.handler_io.input_type and handler.handler_io.input_type.is_void:
                inp = {}
            else:
                inp = InputPayload(
                    required=False,
                    contentType=handler.handler_io.accept,
                    jsonSchema=json_schema_from_type_hint(handler.handler_io.input_type),
                )
            # output
            if handler.handler_io.output_type and handler.handler_io.output_type.is_void:
                out = OutputPayload(setContentTypeIfEmpty=False)
            else:
                out = OutputPayload(
                    setContentTypeIfEmpty=False,
                    contentType=handler.handler_io.content_type,
                    jsonSchema=json_schema_from_type_hint(handler.handler_io.output_type),
                )
            # add the handler
            service_handlers.append(
                Handler(
                    name=handler.name,
                    ty=ty,
                    input=inp,
                    output=out,
                    description=handler.description,
                    metadata=handler.metadata,
                    inactivityTimeout=int(handler.inactivity_timeout.total_seconds() * 1000)
                    if handler.inactivity_timeout
                    else None,
                    abortTimeout=int(handler.abort_timeout.total_seconds() * 1000) if handler.abort_timeout else None,
                    journalRetention=int(handler.journal_retention.total_seconds() * 1000)
                    if handler.journal_retention
                    else None,
                    idempotencyRetention=int(handler.idempotency_retention.total_seconds() * 1000)
                    if handler.idempotency_retention
                    else None,
                    workflowCompletionRetention=int(handler.workflow_retention.total_seconds() * 1000)
                    if handler.workflow_retention
                    else None,
                    enableLazyState=handler.enable_lazy_state,
                    ingressPrivate=handler.ingress_private,
                    retryPolicyInitialInterval=int(
                        handler.invocation_retry_policy.initial_interval.total_seconds() * 1000
                    )
                    if handler.invocation_retry_policy and handler.invocation_retry_policy.initial_interval
                    else None,
                    retryPolicyMaxInterval=int(handler.invocation_retry_policy.max_interval.total_seconds() * 1000)
                    if handler.invocation_retry_policy and handler.invocation_retry_policy.max_interval
                    else None,
                    retryPolicyMaxAttempts=int(handler.invocation_retry_policy.max_attempts)
                    if handler.invocation_retry_policy and handler.invocation_retry_policy.max_attempts is not None
                    else None,
                    retryPolicyExponentiationFactor=float(handler.invocation_retry_policy.exponentiation_factor)
                    if handler.invocation_retry_policy
                    and handler.invocation_retry_policy.exponentiation_factor is not None
                    else None,
                    retryPolicyOnMaxAttempts=(
                        handler.invocation_retry_policy.on_max_attempts.upper()
                        if handler.invocation_retry_policy
                        and handler.invocation_retry_policy.on_max_attempts is not None
                        else None
                    ),
                )
            )
        # add the service
        description = service.service_tag.description
        metadata = service.service_tag.metadata
        services.append(
            Service(
                name=service.name,
                ty=service_type,
                handlers=service_handlers,
                description=description,
                metadata=metadata,
                inactivityTimeout=int(service.inactivity_timeout.total_seconds() * 1000)
                if service.inactivity_timeout
                else None,
                abortTimeout=int(service.abort_timeout.total_seconds() * 1000) if service.abort_timeout else None,
                journalRetention=int(service.journal_retention.total_seconds() * 1000)
                if service.journal_retention
                else None,
                idempotencyRetention=int(service.idempotency_retention.total_seconds() * 1000)
                if service.idempotency_retention
                else None,
                enableLazyState=service.enable_lazy_state if isinstance(service, (Workflow, VirtualObject)) else None,
                ingressPrivate=service.ingress_private,
                retryPolicyInitialInterval=int(service.invocation_retry_policy.initial_interval.total_seconds() * 1000)
                if service.invocation_retry_policy and service.invocation_retry_policy.initial_interval
                else None,
                retryPolicyMaxInterval=int(service.invocation_retry_policy.max_interval.total_seconds() * 1000)
                if service.invocation_retry_policy and service.invocation_retry_policy.max_interval
                else None,
                retryPolicyMaxAttempts=int(service.invocation_retry_policy.max_attempts)
                if service.invocation_retry_policy and service.invocation_retry_policy.max_attempts is not None
                else None,
                retryPolicyExponentiationFactor=float(service.invocation_retry_policy.exponentiation_factor)
                if service.invocation_retry_policy and service.invocation_retry_policy.exponentiation_factor is not None
                else None,
                retryPolicyOnMaxAttempts=(
                    service.invocation_retry_policy.on_max_attempts.upper()
                    if service.invocation_retry_policy and service.invocation_retry_policy.on_max_attempts is not None
                    else None
                ),
            )
        )

    if endpoint.protocol:
        protocol_mode = PROTOCOL_MODES[endpoint.protocol]
    else:
        protocol_mode = PROTOCOL_MODES[discovered_as]
    return Endpoint(protocolMode=protocol_mode, minProtocolVersion=5, maxProtocolVersion=5, services=services)
